/*!**************************************************************************
*!
*! FILE NAME  : init.c
*!
*! DESCRIPTION: Implements a mini init daemon.
*!
*!---------------------------------------------------------------------------
*! HISTORY
*!
*! $Log: init.c,v $
*! Revision 1.51  2004-03-26 15:25:42  johanna
*! Start signal-handler after handling rc.init.
*!
*! Revision 1.50  2004/01/27 10:32:49  johanna
*! Handle failing waitpid due to signal.
*!
*! Revision 1.49  2003/11/24 08:45:28  henriken
*! Fixed a sigchild race that could cause init to hang for tasks with
*! action set to wait.
*! Made perror log source of message.
*!
*! Revision 1.48  2003/11/06 15:09:36  henriken
*! (catch_child): No longer needed to re-install signal handler.
*! (init_signal_handling): Install SIGCHLD handler with sigaction.  Do not
*! save the old signal action since it isn't used anyway.
*! (handle_processes): Block SIGCHLD during fork to avoid a race.
*!
*! Revision 1.47  2003/07/11 13:25:29  jonashg
*! Don't complain about missing rc.init since it's not required.
*!
*! Revision 1.46  2003/02/05 13:54:14  starvik
*! Only fake reboot if FAKE_REBOOT is set
*!
*! Revision 1.45  2002/12/06 14:45:41  jonashg
*! * Run /sbin/rc.init if it exists.
*! * perror messages reveals origin.
*!
*! Revision 1.44  2002/11/07 10:59:30  jonashg
*! Set RUNLEVEL and PREVLEVEL environment variables a la sysvinit.
*!
*! Revision 1.43  2002/11/07 08:34:30  jonashg
*! Implemented support for the sysinit action.
*!
*! Revision 1.42  2002/09/05 13:01:26  aurer
*! * Each spawned process is the session leader of a new session.
*! * Signals are sent to all processes in the process group of the session
*!   leader.
*!
*! Revision 1.41  2002/08/15 16:24:11  sebsjo
*! Only call print_proc_elem() if INITD_DEBUG is defined.
*!
*! Revision 1.40  2002/07/03 15:01:45  sebsjo
*! Init now reboots with newer versions of uClibc.
*!
*! Revision 1.39  2001/12/20 15:46:58  johana
*! Added #include <sys/time.h>
*!
*! Revision 1.38  2001/12/18 14:56:20  jonashg
*! Implemented the -t option.
*!
*! Revision 1.37  2001/12/18 13:46:11  jonashg
*! Use select timeout instead of sleep for faster reaction to fifo requests.
*!
*! Revision 1.36  2001/11/16 13:31:37  jonashg
*! * Keep FIFO open 24/7 in parent and close FIFO in child instead.
*!   (elinux can handle close between vfork and execve).
*! * Don't start processes with KILLME flag set and clear KILLME flag when
*!   started (this is tricky business and I was convinced that it worked
*!   before, and even more so now :).
*!
*! Revision 1.35  2001/11/15 17:51:44  jonashg
*! * Use FIFO (/dev/initctl) instead of file (/tmp/init.switch).
*! * Keep fifo open as much as possible to avoid rootfs reads (needed when
*!   upgrading root partition). FIFO is only closed when forking so that
*!   children doesn't inherit the file descriptor.
*!
*! Revision 1.34  2001/10/04 15:03:30  jonashg
*! Remove all whitespaces instead of just linefeed and space.
*!
*! Revision 1.33  2001/10/04 14:57:07  jonashg
*! Set PATH if not set (a la sysvinit).
*!
*! Revision 1.32  2001/09/21 12:09:56  johana
*! Remove trailing spaces from inittab entries.
*!
*! Revision 1.31  2001/05/28 11:33:23  jonashg
*! Tabs to spaces :)
*!
*! Revision 1.30  2001/05/28 11:30:23  jonashg
*! Added handling of the inittab wait action.
*!
*! Revision 1.29  2001/05/28 08:56:05  jonashg
*! Cleanup: fixed the indentation mess (no more tabs, please).
*!
*! Revision 1.28  2001/05/28 08:45:15  jonashg
*! Check that processes that should be stopped when changing runlevel really dies
*! and send SIGKILL to those that don't.
*!
*! Revision 1.27  2001/05/18 15:18:56  jonashg
*! Start processes that should be started for a runlevel when switching to
*! that runlevel. Use global runlevel variable so that the signal handler can know
*! if the process died because of runlevel change and not assume it mysteriously
*! died (and maybe should be respawned).
*!
*! (Before this change no processes at all was marked for start when
*! switching runlevel, but those with action "respawn" happened to be restarted
*! when switching to one of its runlevels because the were marked for restart
*! when they were stopped. Well, never mind, it should work now anyway.)
*!
*! Revision 1.26  2001/05/18 11:44:05  jonashg
*! * Don't allocate list elements for empty lines in inittab!
*! * Don't allocate an extra empty element in list.
*!
*! Action "once" still doesn't work when entering the runlevel a second time.
*! I will look into this later.
*! (The parsing of inittab is still ugly.)
*!
*! Revision 1.25  2001/05/10 15:25:59  hp
*! Make reboot pass proper arguments for glibc function
*!
*! Revision 1.24  2001/03/28 14:52:49  jonashg
*! Reboot works the same way on LX.
*!
*! Revision 1.23  2001/01/17 13:17:37  johana
*! Made -r work again.
*! BTW: The indentation in this file is a mess, mixing tabs and 2 space.
*!
*! Revision 1.22  2001/01/10 10:35:41  jonashg
*! Print error output if execution of rcfile fails.
*!
*! Revision 1.21  2000/09/06 06:38:02  johana
*! Increased max nbr of arguments (20) and row length (256), removed some warnings.
*!
*! Revision 1.20  2000/09/01 16:11:19  finn
*! Cleanup.
*!
*! Revision 1.19  2000/08/14 12:25:42  torbjore
*! Made it possible to reboot using init.
*!
*! Revision 1.18  2000/06/22 09:31:43  sebsjo
*! Fixed a bug introduced by me
*!
*! Revision 1.17  2000/06/08 16:42:16  finn
*! Cleanup. Again.
*!
*! Revision 1.16  2000/06/08 15:40:44  sebsjo
*! When switching runlevels processes marked as once will preserve their status
*!
*! Revision 1.15  2000/06/08 15:30:56  finn
*! Cleanup.
*!
*! Revision 1.14  2000/04/07 14:22:40  macce
*! Running init manually with runlevel as an argument trigs process 1 to switch runlevel.
*!
*! Revision 1.13  2000/03/06 10:38:08  johana
*! Added LICENSE info (GPL)
*!
*! Revision 1.12  2000/02/18 01:37:56  bjornw
*! * Rewrote read_inittab to use buffered input (fgets) instead of read-loops
*!   (fixed some bugs and a mysterious SIGSEGV in host-build)
*! * Accepts the -r <root> option, which chroot's to the selected dir and
*!   skips running the rc.init script. Used in host-runs.
*!
*! Revision 1.11  1999/12/08 15:36:56  bjornw
*! * Use vfork() instead of fork() in ELINUX
*! * Dont sleep in child before execv
*! * Sleep 0.25s in parent between starting daemons, not 1s
*!
*! Revision 1.10  1999/11/10 15:17:52  johana
*! Don't do strsep on the last (process) inittab parameter since it screws up
*! any colons in parameters (like http://) + some clean up
*!
*! Revision 1.9  1999/11/08 16:19:20  pkj
*! Removed some compiler warnings.
*!
*! Revision 1.8  1999/11/05 13:50:14  torbjore
*! * some changes in switching runlevels
*!
*! Revision 1.7  1999/10/26 06:55:20  torbjore
*! * altered kill sequence in flash upgrade
*!
*! Revision 1.6  1999/10/25 21:59:53  torbjore
*! * checks for ftp flash upgrade
*!
*! Revision 1.5  1999/10/22 12:58:33  macce
*! rc.init has moved to /usr/rc.d/
*!
*! Revision 1.4  1999/10/01 13:05:53  bjornw
*! Run the /etc/rc.init script at start, before executing inittab.
*!
*! Revision 1.3  1999/09/03 08:51:38  tobiasa
*! Removed warnings
*!
*! Revision 1.2  1999/07/19 16:42:17  bjornw
*! * Handle SIGCHLD using a real signal handler. It isn't possible to SIGIGN
*!   SIGCHLD and at the same time wait4 children to get their exit code.
*!
*! * If execv() fails, exit with I_EXEC_FAILED exit code, so that init will
*!   realise the task cannot be started, which will remove the I_VALID bit of
*!   the task. This is possibly not the behaviour we want, but it's better
*!   than trying to keep exec'ing non-existant files every second..
*!
*! * Sleep(1) in parent and child. In parent to slow things down, in child to
*!   give the parent time to store the pid value..
*!
*! * Globalized proc_elem_list from main() because it's needed by the Reaper.
*!
*!
*! DATE         NAME               CHANGES
*! ----         ----               -------
*! Jun 28 1999  Torbjorn Eliasson  Initial version
*! Jul 05 1999  Torbjorn Eliasson  Added runlevel switch
*! Jul 06 1999  Torbjorn Eliasson  Added pid file
*!
*!---------------------------------------------------------------------------
*!
*! (C) Copyright 1999, 2000, Axis Communications AB, LUND, SWEDEN
*!
*! This file is distributed under the Gnu Public License (GPL),
*! please see the file LICENSE for further information.
*!
*! $Id: init.c,v 1.51 2004-03-26 15:25:42 johanna Exp $
*!
*!**************************************************************************/

#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

//#define DEBUG
//#define I_DEBUG
//#define INITD_DEBUG

#define RCFILE "rc.init"
#define INIT_FIFO "/dev/initctl"
#define INIT_MAGIC 0x20011115

#ifdef DEBUG
#define D(x) x
#else
#define D(x)
#endif

#ifdef CREATE_PIDFILE
#define I_WRITEBLE_FS 1
#endif

#define I_MAX_PROC   50
#define I_MAX_LEVEL  7
#define I_MAX_ARGC   20
#define I_ROW_LENGTH 256
#define I_VALID      0x0001
#define I_STARTME    0x0002
#define I_STARTED    0x0004
#define I_KILLME     0x0008  /* Set between TERM and KILL signals. */
#define I_KILLED     0x0010  /* Set between KILL signal and termination. */
#define I_RESPAWN    0x0020
#define I_WAIT       0x0040

#define I_EXIT_REBOOT      -1
#define I_EXIT_HALT        -2
#define I_EXIT_DEF_LEV_MIS -3

#define I_EXEC_FAILED 42

/* Only include <sys/reboot.h> for glibc and newer version of uClibc. Older
 * uClibc versions don't seem to provide a declaration of reboot() anywhere.
 */
#ifdef FAKE_REBOOT
#define reboot(a,b,c) printf("reboot faked\n");
#elif !defined __UCLIBC__ || __UCLIBC__ > 0 || __UCLIBC_MAJOR__ > 9 \
               || (__UCLIBC_MAJOR__ == 9 && __UCLIBC_MINOR__ >= 5)
#include <sys/reboot.h>
#undef reboot
#define reboot(a,b,c) (reboot)(c)
#endif

#define DEFAULT_SLEEPTIME 5
#define PARENT_USLEEP_DELAY 250000
#define RUNLEVELS "0123456"
#define PATH_DFL "PATH=/usr/local/sbin:/sbin:/bin:/usr/sbin:/usr/bin"

struct proc_elem {
  pid_t pid;
  int status;
  char* id;
  char* runlevel;
  char* action;
  char* process;
  int argc;
  char *new_argv[I_MAX_ARGC];
  struct proc_elem* next;
};

struct init_request {
  int magic;
  int sleeptime;
  char runlevel;
};


#ifdef I_DEBUG
static void check_termination(int status);
#endif
#ifdef INITD_DEBUG
static void print_proc_elem(struct proc_elem *p, FILE *s);
#endif
static volatile char runlevel = 0;
static volatile char prevlevel = 'N';
static volatile unsigned char switch_runlevel;
static struct proc_elem *proc_elem_list = NULL;
static int kill_countdown = 0;
static int wait_action_task_done = 0;


/* May be used for a nice exit in debug mode.  */
static void
die (int exit_code)
{
  switch (exit_code)
    {
    case I_EXIT_DEF_LEV_MIS:
      fprintf (stderr, "init: default runlevel is missing\n");
      exit (exit_code);
      break;
    case I_EXIT_REBOOT:
      fprintf (stderr, "init: reboot after runlevel switch\n");
      reboot (0xfee1dead, 672274793, 0x1234567); /* hard reset, now */
      break;
    case I_EXIT_HALT:
      fprintf (stderr, "init: reboot after runlevel switch\n");
      reboot (0xfee1dead, 672274793, 0x1234567); /* hard reset, now */
      break;
    default:
      fprintf (stderr, "init: unknown exit\n");
      reboot (0xfee1dead, 672274793, 0x1234567); /* hard reset, now */
      break;
    }
  exit (exit_code);
}

/* Handles SIGUSR1, forcing a runlevel switch.  */

static void
signal_handler (int sig)
{
  if (sig == SIGUSR1)
    {
      D (fprintf (stderr, "init: switching runlevel\n"));
      switch_runlevel = 1;
    }
}

/* Handles SIGCHLD, checking who died and why. do stuff.  */

static void
catch_child (int sig)
{
  int status;
  int pid;
  struct proc_elem *cur_pp;

  while ((pid = waitpid (-1, &status, WNOHANG)) > 0)
    {
      D (fprintf (stderr, "init: reaping child %d: status %d\n", pid, status));

      /* Find it in our task list.  */

      cur_pp = proc_elem_list;

      while (cur_pp && cur_pp->pid != pid)
        {
          cur_pp = cur_pp->next;
        }

      if (! cur_pp)
        {
          D (fprintf (stderr, "init: weird, caught unknown child die. "
                              "Oh well.\n"));
          continue;
        }

      /* Mark the task as not running.  */

      cur_pp->status &= ~I_STARTED;
      cur_pp->status &= ~I_KILLME;
      cur_pp->status &= ~I_KILLED;

      if (WIFEXITED (status) && WEXITSTATUS (status) == I_EXEC_FAILED)
        {

          /* The child failed its exec call make sure we don't try to
             start it again.  */

          fprintf (stderr, "init: giving up on task %s due to exec failure\n",
                   cur_pp->process);

          cur_pp->status &= ~I_VALID;

          continue;
        }

#ifdef I_WRITEBLE_FS
      do_update = 1;
#endif
#ifdef I_DEBUG
      D (fprintf (stderr, "init: parent found termination of %s (%d)(%d)\n",
                  cur_pp->process, cur_pp->pid, status));
      check_termination (status);
#endif
      /* Check if this task has a wait action. */
      if (cur_pp->status & I_WAIT)
        {
          wait_action_task_done = 1;
          D (fprintf (stderr, "init: --- wait task died %s\n",
                      cur_pp->process));
        }

      /* Check if we should mark the task for restart.  */

      if ((cur_pp->status & I_RESPAWN) && strchr (cur_pp->runlevel, runlevel))
        {
          cur_pp->status |= I_STARTME;
          D (fprintf (stderr, "init: --- marked restart of task %s\n",
                      cur_pp->process));
        }
    }
}


/* Prepares signal handlers to catch SIGCHLD and SIGUSR1.  */

static void
init_signal_handling ()
{
  struct sigaction act;

  sigemptyset (&act.sa_mask);
  act.sa_handler = catch_child;
  act.sa_flags = 0;
  if (sigaction (SIGCHLD, &act, NULL) < 0)
    {
      D (fprintf (stderr,
                  "init: failed to install SIGCHILD signal handler.\n"));
    }

  /* Make sure that SIGUSR1 is catched.  */
  sigemptyset (&act.sa_mask);
  act.sa_handler = signal_handler;
  act.sa_flags = 0;
  if (sigaction (SIGUSR1, &act, NULL) < 0)
    {
      D (fprintf (stderr, "init: failed to install signal handler.\n"));
    }
}


/*

  Opens the inittab file.  Parses every row for a specific pattern defining
  a processes to be started by init.  Every corretly defined process is
  stored in a list returned by this function.

  Layout of inittab

  Every charater following a '#' until end-of-line is ignored.

  process: process should use following row format

  <ID>:<RUNLEVEL>:<ACTION>:<PROCESS>

  <ID> - an uniqe string id for this process definition.

  <RUNLEVEL> - The runlevels used by Elinux are:

  0 - Halt.
  1 - User mode 1
  2 - User mode 2
  3 - User mode 3
  4 - User mode 4
  5 - User mode 5
  6 - Reboot

  <ACTION> - Possible actions are:

  initdefault  - indicates default runlevel for initd
  sysinit      - executed during system boot, runlevels field is ignored
  once         - started once
  wait         - started once and init will wait for its termination
  respawn      - restarted whenever it terminates

  <PROCESS> - process to be started. e.g. /bin/boa -c /etc/httpd

*/

static struct proc_elem *
read_inittab ()
{
  FILE *inittab;
  int row_tok;
  char row_buffer[I_ROW_LENGTH];
  char *row_pos_ptr;
  struct proc_elem *pp = NULL;
  struct proc_elem *cur_pp = NULL;
  struct proc_elem *tmp_pp;

  /* First we need to read inittab.  */
  if (! (inittab = fopen ("/etc/inittab", "r")))
    {
      perror ("init: open /etc/inittab");
      exit (1);
    }

  /* Now we can try to parse every row.  Every valid row is entered
     in database.  */
  D (fprintf (stderr, "init: reading inittab.\n"));

  /* Read each line in the file until EOF.  */

  while (fgets (row_buffer, I_ROW_LENGTH, inittab))
    {
      char *tok;
      int i;

      if (row_buffer[0] == '#')
        {
          continue;  /* skip comments */
        }

      /* Parse row: <ID>:<RUNLEVEL>:<ACTION>:<PROCESS>
       *
       * This really leaks memory, when a line is corrupt
       * we just reloop and don't free the strdup'ed contents.
       * It does not matter very much though since init runs
       * only once.
       */
      row_pos_ptr = row_buffer;
      /* id */
      if (! (tok = strsep (&row_pos_ptr, ":")) || ! row_pos_ptr)
        {
          continue;
        }
      if (! (tmp_pp = (struct proc_elem*) malloc (sizeof (struct proc_elem))))
        {
          perror ("init: malloc");
          fclose (inittab);
          return 0;
        }
      memset (tmp_pp, 0, sizeof *tmp_pp);
      if (!pp)
        {
          /* Remember the head element of list (pp). */
          pp = cur_pp = tmp_pp;
        }
      else
        {
          cur_pp->next = tmp_pp;
          cur_pp = cur_pp->next;
        }
      cur_pp->id = strdup (tok);
      /* runlevel */
      if (! (tok = strsep (&row_pos_ptr, ":")))
        {
          continue;
        }
      cur_pp->runlevel = strdup (tok);
      /* action */
      if (! (tok = strsep (&row_pos_ptr, ":")))
        {
          continue;
        }
      cur_pp->action = strdup (tok);

      i = strlen (row_pos_ptr);
      /* Remove all trailing whitespaces */
      while (i && isspace(row_pos_ptr[i - 1]))
        {
          row_pos_ptr[i - 1] = 0;
          i--;
        }

      if (*row_pos_ptr)
        {
          cur_pp->process = strdup (row_pos_ptr);
        }
      else
        {
          cur_pp->process = NULL;
        }

      cur_pp->status |= I_VALID;
    }

  fclose (inittab);

  /* Check entries.  */

  cur_pp = pp;
  while (cur_pp != NULL)
    {
      if (cur_pp->status & I_VALID)
        {
          /* Temporary remove the valid flag.  */
          cur_pp->status &= ~I_VALID;

          /* Find default runlevel.  */
          if (! strncmp (cur_pp->action, "initdefault", 11))
            {
              if (strlen (cur_pp->runlevel) == 1)
                {
                  if ((*cur_pp->runlevel > '0') ||
                      (*cur_pp->runlevel > '7'))
                    {
                      runlevel = *cur_pp->runlevel;
                      D (fprintf (stderr, "init: default runlevel is %c\n",
                                  runlevel));
                    }
                }

              /* Don't process this any further.  */
              cur_pp = cur_pp->next;
              continue;
            }

          /* Check if correct action. */
          if (! strncmp (cur_pp->action, "sysinit", 7))
            {
              D (fprintf (stderr, "init: %s - defined for sysinit\n",
                          cur_pp->id));
              cur_pp->status |= I_VALID | I_WAIT | I_STARTME;
              cur_pp->runlevel = "*"; /* Not used */
            }
          if (! strncmp (cur_pp->action, "once", 4))
            {
              D (fprintf (stderr, "init: %s - defined for once\n", cur_pp->id));
              cur_pp->status |= I_VALID | I_STARTME;
            }
          if (! strncmp (cur_pp->action, "wait", 4))
            {
              D (fprintf (stderr, "init: %s - defined for wait\n", cur_pp->id));
              cur_pp->status |= I_VALID | I_WAIT | I_STARTME;
            }
          if (! strncmp (cur_pp->action, "respawn", 7))
            {
              D (fprintf (stderr, "init: %s - defined for respawn\n",
                          cur_pp->id));
              cur_pp->status |= I_VALID | I_RESPAWN | I_STARTME;
            }

          /* Create argvs.  */
          row_tok = 0;
          row_pos_ptr = cur_pp->process;
          for (row_tok = 0; row_tok < I_MAX_ARGC; row_tok++)
            {
              cur_pp->new_argv[row_tok] = strsep (&row_pos_ptr, " ");
              if (cur_pp->new_argv[row_tok])
                {
                  cur_pp->argc++;
                }
              else
                {
                  break;
                }
            }
#ifdef INITD_DEBUG
          /* Show entry.  */
          print_proc_elem (cur_pp, stderr);
#endif

          /* Move sysinits to head of list so they get started first. */
          if (! strncmp (cur_pp->action, "sysinit", 7))
            {
              if (cur_pp != pp)
                {
                  /* Simple linear search. */
                  tmp_pp = pp;
                  while (tmp_pp->next && (tmp_pp->next != cur_pp))
                    {
                            tmp_pp = tmp_pp->next;
                    }

                  if (tmp_pp->next == cur_pp)
                    {
                      /* Move cur_pp to head. */
                      tmp_pp->next = cur_pp->next;
                      cur_pp->next = pp;
                      pp = cur_pp;

                      /* Make sure we continue at the next element below. */
                      cur_pp = tmp_pp;
                    }
                }
            }
        }
      /* Step forward.  */
      cur_pp = cur_pp->next;
    }

  return pp;
}


/* Print the contents of a node.  */
#ifdef INITD_DEBUG
static void
print_proc_elem (struct proc_elem *p, FILE *s)
{
  int i;

  fprintf (s, "struct proc_elem: 0x%p\n", p);
  fprintf (s, "{\n");
  fprintf (s, "        %d, /* pid */\n", p->pid);
  fprintf (s, "        %d, /* status */\n", p->status);
  fprintf (s, "            %s %s %s %s %s %s %s",
           p->status & I_VALID? "I_VALID":"",
           p->status & I_STARTME? "I_STARTME":"",
           p->status & I_STARTED? "I_STARTED":"",
           p->status & I_KILLME? "I_KILLME":"",
           p->status & I_KILLED? "I_KILLED":"",
           p->status & I_RESPAWN? "I_RESPAWN":"",
           p->status & I_WAIT? "I_WAIT":"");
  fprintf (s, "        \"%s\", /* id */\n", p->id);
  fprintf (s, "        \"%s\", /* runlevel */\n", p->runlevel);
  fprintf (s, "        \"%s\", /* action */\n", p->action);
  fprintf (s, "        \"%s\", /* process */\n", p->process);
  fprintf (s, "        %d, /* argc */\n", p->argc);
  fprintf (s, "        /* arguments: */\n");

  for (i = 0; i < p->argc; i++)
    {
      fprintf (s, "         \"%s\" /* %d */\n", p->new_argv[i], i);
    }

  fprintf(s, "}\n");
}
#endif

/* Printout of process element list.  */

#ifdef I_DEBUG
static void
check_list (struct proc_elem *pp)
{
  int row_tok;
  struct proc_elem *cur_pp = pp; 

  while (cur_pp != NULL)
    {
      if (cur_pp->status & I_VALID)
        {
          fprintf (stderr, "%s - %X %s %s %s \n",
                   cur_pp->id,
                   cur_pp->status,
                   cur_pp->runlevel,
                   cur_pp->action,
                   cur_pp->process);
          for (row_tok = 0; row_tok < I_MAX_ARGC; row_tok++)
            {
              fprintf (stderr, "(%s)\n", cur_pp->new_argv[row_tok]);
            }
        }
      cur_pp = cur_pp->next;
    }
}
#endif


/* Checks the reson to why a process terminated.  Used only in debug.  */

#ifdef I_DEBUG
static void
check_termination (int status)
{
  if (WIFEXITED (status))
    {
      D (fprintf (stderr, "init: WIFEXITED %d\n", status));
      D (fprintf (stderr, "init: WEXITSTATUS %d\n", WEXITSTATUS (status)));
    }
  else if (WIFSIGNALED (status))
    {
      D (fprintf (stderr, "init: WIFSIGNALED %d\n", WIFSIGNALED (status)));
      D (fprintf (stderr, "init: WTERMSIG %d\n", WTERMSIG (status)));
    }
  else if (WIFSTOPPED (status))
    {
      D (fprintf (stderr, "init: WIFSTOPPED %d\n", WIFSTOPPED (status)));
      D (fprintf (stderr, "init: WSTOPSIG %d\n", WSTOPSIG (status)));
    }
  else
    {
      D (fprintf (stderr, "init: termination unknown\n"));
    }
}
#endif

/* Creates a file containing all PID's.  */

#ifdef I_WRITEBLE_FS
static void
update_pid_file (struct proc_elem *pp)
{
  int fd;
  int i;
  char buffer[80];
  struct proc_elem *cur_pp;

  errno = 0;
  if ((fd = open (".pidtab", O_CREAT | O_TRUNC | O_WRONLY,
                  S_IROTH | S_IRGRP | S_IWUSR | S_IRUSR)) < 0)
    {
      D (fprintf (stderr, "init: error creating .pidtab %d\n\n", errno));
      return;
    }

  /* Fill file with pids.  */

  sprintf (buffer, "     init:%6d:init\n", getpid ());
  write (fd, buffer, strlen (buffer));
  cur_pp = pp;

  while (cur_pp)
    {
      if (cur_pp->status & I_VALID)
        {
          if (cur_pp->id)
            {
              sprintf (buffer, "%9s:%6d:", cur_pp->id, cur_pp->pid);
            }
          else
            {
              sprintf (buffer, ":%6d:", cur_pp->pid);
            }
          write (fd, buffer, strlen (buffer));
          for (i = 0; i < I_MAX_ARGC; i++)
            {
              if (cur_pp->new_argv[i] == NULL)
                {
                  break;
                }
              sprintf (buffer, "%s ", cur_pp->new_argv[i]);
              write (fd, buffer, strlen (buffer));
            }
          write (fd, "\n", 1);
        }
      cur_pp = cur_pp->next;
    }
  fsync (fd);
  close (fd);
}
#endif


/* Start processes to be started for default runlevel.  Check that no
   processes marked as ``respawn'' has died.  */

static void
handle_processes (struct proc_elem *pp)
{
  int fifo_fd;
  char new_level;
  int count_down = 0;
#ifdef I_WRITEBLE_FS
  int do_update;
#endif
  struct proc_elem *cur_pp;
  fd_set fds;
  struct timeval tv;
  int fifo_activity;
  int sleeptime = DEFAULT_SLEEPTIME; /* Just to get rid of compiler warning. */
  static int sysinit = 1; /* True during boot. */
  char env_runlevel[] = "RUNLEVEL=x";
  char env_prevlevel[] = "PREVLEVEL=x";

  if ((fifo_fd = open (INIT_FIFO, O_RDWR | O_NONBLOCK)) < 0)
    {
      perror("init: open " INIT_FIFO);
      fprintf (stderr, "init: runlevel change disabled!\n");
    }

  /* Starting processes for the default level.  */

  D (fprintf (stderr, "init: starting processes at level : %c\n", runlevel));

  /* Start of inifinite action loop.  */

#ifdef I_WRITEBLE_FS
  do_update = 1;
#endif
  for (;;)
    {
      /* Check every child process if there is any action to be taken.  */
      errno = 0;
      cur_pp = pp;
      while (cur_pp)
        {
          if (! strchr (cur_pp->runlevel, runlevel)
              && cur_pp->status & I_VALID
              && cur_pp->status & I_STARTED
              && ! (cur_pp->status & I_STARTME)
              && ! (cur_pp->status & I_KILLME)
              && ! (cur_pp->status & I_KILLED))
            {
              /* Found process to terminate. */
              fprintf (stderr, "init: sending process group %d (%s) the "
                               "TERM signal.\n", cur_pp->pid, cur_pp->id);
              kill_countdown = sleeptime;
              kill (-cur_pp->pid, SIGTERM);
              cur_pp->status |= I_KILLME;
            }

          if ((cur_pp->status & I_STARTED)
              && (cur_pp->status & I_VALID)
              && (cur_pp->status & I_KILLME)
              && (kill_countdown <= 0))
            {
                /* Found process that didn't terminate on time. KILL it! */
                fprintf (stderr, "init: sending process group %d (%s) the "
                                 "KILL signal.\n", cur_pp->pid, cur_pp->id);
                kill (-cur_pp->pid, SIGKILL);
                cur_pp->status |= I_KILLED;

                /* Hope it dies and don't try any more. */
                cur_pp->status &= ~I_KILLME;
            }

          if ((strchr (cur_pp->runlevel, runlevel) ||
               (sysinit && (strncmp (cur_pp->action, "sysinit", 7) == 0)))
              && cur_pp->status & I_STARTME
              && cur_pp->status & I_VALID
              && ! (cur_pp->status & I_STARTED))
            {
              sigset_t sigset;
              int sigerr;
              /* Found process to start.  */

              /* Since all sysinits are at the beginning of the list we now
               * know if they all have been processed and can prevent them
               * from being started in the future.
               */
              if (sysinit && (strncmp (cur_pp->action, "sysinit", 7) != 0))
                {
                  sysinit = 0; /* Never ever true again. */
                }

#ifdef I_WRITEBLE_FS
              do_update = 1;
#endif
              cur_pp->status |= I_STARTED;
              cur_pp->status &= ~I_STARTME;
              cur_pp->status &= ~I_KILLME;

#ifdef __elinux__
              cur_pp->pid = vfork ();
#else
              if ((sigerr = sigemptyset (&sigset)) != 0)
                {
                  perror ("init: sigemptyset");
                }
              else if ((sigerr = sigaddset (&sigset,
                                            SIGCHLD)) != 0)
                {
                  perror ("init: sigaddset");
                }
              else if ((sigerr = sigprocmask (SIG_BLOCK,
                                              &sigset,
                                              NULL)) != 0)
                {
                  perror ("init: sigprocmask");
                }
              D (if (sigerr != 0) 
                 {
                   fprintf (stderr, 
                            "init: failed SIGCHILD block prior to fork.\n");
                 });

              cur_pp->pid = fork ();
#endif

              if (cur_pp->pid)
                {
                  D (fprintf (stderr, "init: started process %s, pid %d\n",
                              cur_pp->id, cur_pp->pid));

#ifndef __elinux__
                  if ((sigerr = sigprocmask (SIG_UNBLOCK, &sigset, NULL)) != 0)
                    {
                      fprintf (stderr, 
                               "init: parent: failed SIGCHILD "
                               "unblock post fork.\n");
                    }
#endif
                  if (cur_pp->status & I_WAIT)
                    {
                      while (!wait_action_task_done)
                        {
                          pause ();
                        }
                      wait_action_task_done = 0;
                      cur_pp->status &= ~I_STARTED;
                    }
                  else
                    {
                      /* Some room between starts.  */
                      usleep (PARENT_USLEEP_DELAY);
                    }
                }
              else
                {
#ifndef __elinux__
                  if ((sigerr = sigprocmask (SIG_UNBLOCK, &sigset, NULL)))
                  {
                      fprintf (stderr, "init: child: failed SIGCHILD "
                               "unblock post fork.\n");
                    }
#endif
                  /* We don't want new processes to have fifo_fd. */
                  if (close(fifo_fd) < 0)
                    {
                      perror("init: close " INIT_FIFO);
                    }

                  /* Create a session. */
                  setsid();

                  /* Set environment variables. */
                  env_runlevel[9] = (sysinit)? 'S' : runlevel;
                  putenv(env_runlevel);
                  env_prevlevel[10] = prevlevel;
                  putenv(env_prevlevel);

                  D (fprintf (stderr, "init: child %s %d started\n",
                              cur_pp->process, getpid ()));

                  D (fprintf (stderr, "init: starting process %s\n",
                              cur_pp->process));
                  execv (cur_pp->new_argv[0], cur_pp->new_argv);

                  D (fprintf (stderr, "init: execv returned errno = %d \n",
                              errno));

                  /* Exec failed.  We exit with a code that the parent can
                     recognize.  */

                  exit (I_EXEC_FAILED);
                }
            }

          /* Step forward.  */
          cur_pp = cur_pp->next;
        }

#ifdef I_WRITEBLE_FS
      if (do_update)
        {
          update_pid_file (pp);
          do_update = 0;
        }
#endif

      FD_ZERO (&fds);
      FD_SET (fifo_fd, &fds);
      memset (&tv, 0, sizeof(tv));
      tv.tv_sec = 1; /* Must be 1 so the -t option is in seconds. */
      if ((fifo_activity = select (fifo_fd + 1, &fds, NULL, NULL, &tv)) < 0)
        {
          if (errno != EINTR) /* Don't call perror on SIGCHLD. */
            {
              perror ("init: select " INIT_FIFO);
            }
          continue;
        }
      if (! fifo_activity && (kill_countdown > 0))
        {
          kill_countdown--;
        }
      if (FD_ISSET(fifo_fd, &fds))
        {
          int size;
          struct init_request request;

          if ((size = read (fifo_fd, &request, sizeof (request))) < 0)
            {
              if (errno != EAGAIN)
                {
                  perror ("init: read " INIT_FIFO);
                }
              continue;
            }

          if ((size != sizeof (request)) || (request.magic != INIT_MAGIC))
            {
              fprintf (stderr, "init: ignoring bogus reguest on " INIT_FIFO
                               "\n");
              continue;
            }

          sleeptime = request.sleeptime;
          new_level = request.runlevel;
          if (strchr (RUNLEVELS, new_level)
              && (new_level != runlevel))
            {
              /* New level ok.  */
              prevlevel = runlevel;
              runlevel = new_level;
              switch_runlevel = 1;

              /* If runlevel is 0 then HALT.
                 If runlevel is 6 then REBOOT.   */
              if (runlevel == '0' || runlevel == '6')
                {
                  count_down = 10;
                }
              fprintf (stderr, "init: switching to runlevel %c.\n", runlevel);
            }
        }

      if (switch_runlevel)
        {
          D (fprintf (stderr, "init: handle switch\n"));
          if (count_down > 0)
            {
              if (count_down-- < 2)
                {

                  /* If runlevel is 0 then HALT.  */
                  if (runlevel == '0')
                    {
                      die (I_EXIT_HALT);
                    }

                  /* If runlevel is 6 then REBOOT.  */
                  if (runlevel == '6')
                    {
                      die (I_EXIT_REBOOT);
                    }
                }
            }
          else
            {
              struct proc_elem *tmp_pp = pp;
              while (tmp_pp)
                {
                  if (strchr (tmp_pp->runlevel, runlevel) &&
                      (! (tmp_pp->status & I_STARTED) ||
                       (tmp_pp->status & I_KILLME)))
                    {
                      tmp_pp->status |= I_STARTME;
                    }
#ifdef INITD_DEBUG 
                  print_proc_elem (tmp_pp, stderr);
#endif
                  tmp_pp = tmp_pp->next;
                }
              switch_runlevel = 0;
            }
        }
    }
}

/* Run the /etc/rc.init script if it exists.  */

static void
run_rc_init ()
{
  int pid;

  /* First do the basic check on the scripts existance.  */

  pid = vfork ();

  if (pid == 0)
    {
      static char * const argv[] = {RCFILE, NULL};

      /* In child. Execute the script.  */

      execv ("/sbin/" RCFILE, argv);
      execv ("/usr/rc.d/" RCFILE, argv);

      /* No rc-file found, probably no problem. */
      _exit (0);
    }

  /* Wait for it to finish.  */
  waitpid (pid, NULL, 0);
}

void
usage()
{
  fprintf(stderr, "Usage: init [ -t sec ] [ " RUNLEVELS " ]\n");
  exit (1);
}

/* Call to setup signalhandler, read inittab, start and manage
   processes.  */

int
main (int argc, char **argv)
{
  int no_rc_init = 0;
  int sleeptime = DEFAULT_SLEEPTIME;

  if ((argc > 2) && (! strcmp (argv[1], "-r")))
    {
        /* Do -r stuff later */
    }
  else
    {
      if (getpid () != 1 && (argc < 2 || argc > 4))
        {
          usage ();
        }
      else if (getpid () != 1)
        {
          FILE *file;
          struct init_request request;
          int c;

          while ((c = getopt(argc, argv, "t:")) != EOF)
            {
              switch (c)
                {
                case 't':
                  if (!optarg)
                    {
                      usage ();
                    }
                  sleeptime = atoi(optarg);
                  break;
                default:
                  usage ();
                  break;
                }
            }

          if (optind == argc)
            {
              usage ();
            }

          if (strlen (argv[optind]) != 1 || ! strchr (RUNLEVELS, *argv[optind]))
            {
              usage ();
            }

          request.magic = INIT_MAGIC;
          request.sleeptime = sleeptime;
          request.runlevel = *argv[optind];

          if ((file = fopen (INIT_FIFO, "wb")) == NULL)
            {
              perror ("init: fopen " INIT_FIFO);
              exit (1);
            }
          if (fwrite (&request, sizeof(request), 1, file) != 1)
            {
              fprintf (stderr, "Failed to write request to " INIT_FIFO "!\n");
              exit (1);
            }
          fclose (file);
          exit (0);
      }
    }

  fprintf (stderr, "init: starting up.\n");
  D (fprintf (stderr, "init: parent PID is %d\n\n", getpid ()));

  if (argc > 2)
    {
      if (! strcmp (argv[1], "-r"))
        {
          fprintf(stderr, "init: using chroot(%s)\n", argv[2]);
          if (chroot (argv[2]) < 0)
            {
              perror ("init: chroot");
              exit (1);
            }
          fprintf (stderr, "init: skipping rc.init.\n");
          no_rc_init++;
        }
    }

  if (getenv("PATH") == NULL)
    {
      putenv(PATH_DFL);
    }

  /* Run the /etc/rc.init script.  */
  if (! no_rc_init)
    {
      run_rc_init ();
    }

  /* Setup signal handler.  */
  init_signal_handling ();

  /* Read inittab and create list of process elements.  */
  proc_elem_list = read_inittab ();

  /* Check if default runlevel was specified in inittab.  */
  if (runlevel == 0)
    {
      die (I_EXIT_DEF_LEV_MIS);
    }

  /* Handle process in spec. runlevel.  */
  handle_processes (proc_elem_list);

  /* Should never return.  */
  return 1;
}
